
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging

from enum                       import Enum
from threading                  import Thread, Timer
from datetime                   import datetime, timedelta

from modules.base.utils         import SimpleUtil as SU
from modules.base.mail          import AuraMailer
from modules.core.resources     import ResourceClass
from modules.core.channels      import Channel




class FallbackType(Enum):
    """
    Types of playlists.
    """
    NONE        = { "id": 0, "name": "default", "lqs_sources": [ Channel.QUEUE_A, Channel.QUEUE_A] }    # No fallback active, default playout
    SCHEDULE    = { "id": 1, "name": "schedule", "lqs_sources": [ Channel.FALLBACK_QUEUE_A, Channel.FALLBACK_QUEUE_B]}   # The first played when some default playlist fails
    SHOW        = { "id": 2, "name": "show", "lqs_sources": [ "station_folder", "station_playlist"]}       # The second played when the timeslot fallback fails    
    STATION     = { "id": 3, "name": "station", "lqs_sources": [ "station_folder", "station_playlist"] }    # The last played when everything else fails

    @property
    def id(self):
        return self.value["id"]

    def __str__(self):
        return str(self.value["name"])



class FallbackManager:
    """
    Handles all types of fallbacks in case there is an outage
    for the regular radio programme.

    Attributes:
        config (AuraConfig):        The engine configuration
        logger (AuraLogger):        The logger
        mail   (AuraMailer):        Mail service
        scheduler (AuraScheduler):  The scheduler
    """    
    config = None
    logger = None
    mailer = None
    scheduler = None
    message_timer = None


    
    def __init__(self, config, logger, scheduler, message_timer):
        """
        Constructor

        Args:
            config (AuraConfig):    Holds the engine configuration
        """
        self.config = config
        self.logger = logger
        self.mailer = AuraMailer(self.config)
        self.scheduler = scheduler
        # self.message_timer = message_timer
        self.message_timer = []


    #
    #   PUBLIC METHODS
    #

    def schedule_fallback_playlist(self, schedule, schedule_now=False):
        """
        Evaluates the scheduled fallback and queues it using a timed thread.

        Args:
            schedule_now (Boolean): If `True` it is executed immediately
        """
        timer_start = None
        timer_end = None
        (fallback_type, playlist) = self.get_fallback_playlist(schedule)

        if playlist:
            self.logger.info(f"Resolved {fallback_type.value} fallback")

            def do_schedule(entries):
                self.logger.info(SU.cyan(f"=== set_fallback_playlist('{entries}') ==="))
                self.scheduler.engine.player.start_fallback_playlist(entries)
            def do_unschedule():
                self.logger.info(SU.cyan("=== clear_fallback_playlist() ==="))
                self.scheduler.engine.player.stop_fallback_playlist()

            if schedule_now == True:
                # Update queue immediately
                thread = Thread(target = do_schedule, args = (playlist.entries,))
                thread.start()
            else:          
                # Update queue at the beginning of the timeslot  
                timer_start = FallbackCommandTimer(schedule.start_unix, do_schedule, playlist.entries)
                self.message_timer.append(timer_start)
                timer_start.start()

            # Update fallback channel to be cleared at the end of the timeslot
            timer_end = FallbackCommandTimer(schedule.end_unix, do_unschedule)
            self.message_timer.append(timer_end)
            timer_end.start()
            return (timer_start, timer_end)

        else:
            msg = f"There is no schedule- or show-fallback defined for timeslot#{schedule.schedule_id}. "
            msg += f"The station fallback will be used automatically."
            self.logger.info(msg)



    def resolve_playlist(self, schedule):
        """
        Retrieves the currently planned (fallback) playlist. If a normal playlist is available,
        this one is returned. In case of station fallback no playlist is returned.

        Args:
            schedule (Schedule)
        
        Returns:
            (FallbackType, Playlist)
        """
        planned_playlist = None
        fallback_type = None

        if self.validate_playlist(schedule, "playlist"):
            planned_playlist = schedule.get_playlist()
            fallback_type = FallbackType.NONE
        else:
            (fallback_type, planned_playlist) = self.get_fallback_playlist(schedule)            

        return (fallback_type, planned_playlist)



    def get_fallback_playlist(self, schedule):
        """
        Retrieves the playlist to be used in a fallback scenario.

        Args: 
            schedule (Schedule)

        Returns:
            (Playlist)
        """        
        playlist = None
        fallback_type = FallbackType.STATION

        if self.validate_playlist(schedule, "schedule_fallback"):
            playlist = schedule.schedule_fallback[0]
            fallback_type = FallbackType.SCHEDULE
        elif self.validate_playlist(schedule, "show_fallback"):
            playlist = schedule.show_fallback[0]
            fallback_type = FallbackType.SHOW

        return (fallback_type, playlist)



    #
    #   PRIVATE METHODS
    #


    def validate_playlist(self, schedule, playlist_type):
        """
        Checks if a playlist is valid for play-out.

        Following checks are done for all playlists:

            - has one or more entries

        Fallback playlists should either:
                
            - have filesystem entries only
            - reference a recording of a previous playout of a show (also filesystem)
        
        Otherwise, if a fallback playlist contains Live or Stream entries,
        the exact playout behaviour can hardly be predicted.
        """
        playlist = getattr(schedule, playlist_type)
        if playlist \
            and isinstance(playlist, list) \
            and playlist[0].entries \
            and len(playlist[0].entries) > 0:

            # Default playlist
            if playlist_type == "playlist":
                return True

            # Fallback playlist
            elif playlist[0].entries:
                is_fs_only = True
                for entry in playlist[0].entries:
                    if entry.get_content_type() not in ResourceClass.FILE.types:
                        self.logger.error(SU.red("Fallback playlist of type '%s' contains not only file-system entries! \
                            Skipping fallback level..." % playlist_type))
                        is_fs_only = False
                        break                
                return is_fs_only

        return False





class EngineCommandTimer(Timer):
    """
    Timer for timed executing of Engine commands.
    """
    timer_store = {}    
    logger = logging.getLogger("AuraEngine")
    timer_id = None
    timer_type = None
    param = None
    diff = None
    dt = None


    def __init__(self, timer_type="BASE", due_time=None, func=None, param=None):
        """
        Constructor
        """
        now_unix = Engine.engine_time()        
        self.timer_type = timer_type       
        self.timer_id = f"{timer_type}:{func.__name__}:{due_time}"
                
        diff = due_time - now_unix
        if diff < 0.0:
            msg = f"Trying to create timer in the past: {self.timer_id}"
            self.logger.error(SU.red(msg))
            raise Exception(msg)

        self.diff = diff
        self.dt = datetime.now() + timedelta(seconds=diff)
        self.func = func
        self.param = param

        def wrapper_func(param=None):

            # Remove from cache
            self.logger.info(SU.green(f"Removing old timer with ID: {self.timer_id}"))
            del EngineCommandTimer.timer_store[self.timer_id]            
            # Call actual function
            if param: func(param,) 
            else: func()

        Timer.__init__(self, diff, wrapper_func, (param,))       
        self.update_cache()
        self.logger.info(SU.green(f"Created command timer with ID: {self.timer_id}"))
        

    
    def update_cache(self):
        """
        Adds the instance to the cache and cancels any previously existing commands.
        """
        existing_command = None
        if self.timer_id in EngineCommandTimer.timer_store:
            existing_command = EngineCommandTimer.timer_store[self.timer_id]    
        if existing_command:
            self.logger.info(SU.green(f"Cancelling previous timer with ID: {self.timer_id}"))
            existing_command.cancel()
        EngineCommandTimer.timer_store[self.timer_id] = self



    def print_active_timers(self):
        """
        Prints a list of active timers to the log.
        """
        for id, timer in EngineCommandTimer.timer_store.values():
            EngineCommandTimer.logger.info(str(timer))



    def __str__(self):
        """
        String represenation of the timer.
        """
        return f"[{self.timer_id}] COMMAND TIMER due at {str(self.dt)} (alive: {self.is_alive()})"





class FallbackCommandTimer(EngineCommandTimer):
    """
    Timer for executing timed scheduling of fallback playlists.
    """
    def __init__(self, diff=None, func=None, param=None):
        """
        Constructor
        """        
        super().__init__("FALLBACK", diff, func, param)
        self.logger.info("Executing scheduled fallback playlist update '%s' in %s seconds..." % \
            (str(func.__name__), str(diff)))

